from lazilypy import Lazy

from conftest import (
    get_lazy,
    lazy_created_and_not_started,
    lazy_builder_called_and_instance_started,
)


def test_addition():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int + 2 == 3
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 + lazy_int == 3
    lazy_builder_called_and_instance_started(lazy_int)

    assert get_lazy(int, 1) + get_lazy(int, 1) == 2


def test_concat():
    lazy_one = get_lazy(str, "one")
    lazy_created_and_not_started(lazy_one)
    assert lazy_one + "two" == "onetwo"
    lazy_builder_called_and_instance_started(lazy_one)

    lazy_two = get_lazy(str, "two")
    lazy_created_and_not_started(lazy_two)
    assert "one" + lazy_two == "onetwo"
    lazy_builder_called_and_instance_started(lazy_two)

    assert get_lazy(str, "one") + get_lazy(str, "two") == "onetwo"


def test_contains():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int in [2]) == (1 in [2])
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_list = get_lazy(list, [1, 2])
    lazy_created_and_not_started(lazy_list)
    print(lazy_list)
    assert (1 in lazy_list) == (1 in [1, 2])
    lazy_builder_called_and_instance_started(lazy_list)


def test_true_div():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int / 2 == 1 / 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 / lazy_int == 2 / 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_floor_div():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int // 2 == 1 // 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 // lazy_int == 2 // 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_bitwise_and():
    # TODO: Unsupported
    # lazy_int: int | Lazy = get_lazy(int, 1)
    # lazy_created_and_not_started(lazy_int)
    # assert lazy_int & 2 == 1 & 2
    # lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 & lazy_int == 2 & 1
    assert 2 & lazy_int == 0
    lazy_builder_called_and_instance_started(lazy_int)


def test_bitwise_xor():
    lazy_int: int | Lazy = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int ^ 2 == 1 ^ 2
    assert lazy_int ^ 2 == 3
    lazy_builder_called_and_instance_started(lazy_int)

    # TODO: Unsupported
    # lazy_int = get_lazy(int, 1)
    # lazy_created_and_not_started(lazy_int)
    # assert 2 ^ lazy_int == 2 ^ 1
    # assert 2 ^ lazy_int == 3
    # lazy_builder_called_and_instance_started(lazy_int)


def test_bitwise_inversion():
    lazy_int: int | Lazy = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert ~lazy_int == ~1
    lazy_builder_called_and_instance_started(lazy_int)


def test_bitwise_or():
    lazy_int: int | Lazy = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int | 2 == 1 | 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 | lazy_int == 2 | 1
    assert 2 | lazy_int == 3
    lazy_builder_called_and_instance_started(lazy_int)


def test_power():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int**2 == 1**2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2**lazy_int == 2**1
    lazy_builder_called_and_instance_started(lazy_int)


def test_left_shift():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int << 2 == 1 << 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 << lazy_int == 2 << 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_modulo():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int % 2 == 1 % 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 % lazy_int == 2 % 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_multiplication():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int * 2 == 1 * 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 * lazy_int == 2 * 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_negation():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert -lazy_int == -1
    lazy_builder_called_and_instance_started(lazy_int)


def test_right_shift():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int >> 2 == 1 >> 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 >> lazy_int == 2 >> 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_subtraction():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int - 2 == -1
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 - lazy_int == 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_less():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int < 2) == (1 < 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 < lazy_int) == (2 < 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_less_or_equal():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int <= 2) == (1 <= 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 <= lazy_int) == (2 <= 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_equal():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int == 2) == (1 == 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 == lazy_int) == (2 == 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_not_equal():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int != 2) == (1 != 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 != lazy_int) == (2 != 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_greater_or_equal():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int > 2) == (1 > 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 > lazy_int) == (2 > 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_greater():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int > 2) == (1 > 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 > lazy_int) == (2 > 1)
    lazy_builder_called_and_instance_started(lazy_int)



from logging import Logger
from pprint import pprint
from typing import (
    Generic,
    TypeVar,
    Callable,
    Any,
    Union,
    Tuple,
    Type,
    Optional,
    NewType,
)

from lazilypy import location_info
from lazilypy.defaults import supported_python_operators, proxy_operator, MethodResponse
from lazilypy.logs import logger

_T = TypeVar("_T")

try:

    class MetaNewType(NewType):  # type: ignore[valid-type,misc]
        pass

except TypeError:

    class MetaNewType(object):  # type: ignore[no-redef]
        pass


class Lazy(MetaNewType, Generic[_T]):
    _lazy_ref_instance_: Optional[_T] = None
    _lazy_default_keywords_ = ["__class__", "__supertype__"]
    _lazy_reserved_keywords_: Tuple[str, str] = ("", "")

    def __init__(
        self,
        lazy_cls: Type[_T],
        *args,
        lazy_builder: Callable[[Any], _T] = None,
        lazy_partial: bool = True,
        lazy_logger: Logger = logger,
        **kwargs,
    ):
        # create a copy of default keywords
        self._lazy_default_keywords_ = self._lazy_default_keywords_.copy()
        # Store own class in other reference
        self._lazy_cls_ = self.__class__
        # access to class property from now own should start instance (e.g: class checks)
        self._lazy_default_keywords_.pop(0)
        self._lazy_logger_ = lazy_logger
        # store non-instance class reference
        self._lazy_ref_cls_ = lazy_cls
        self.__supertype__ = lazy_cls
        # defaults builder to cls __init__
        self._lazy_builder_ = lazy_builder if lazy_builder is not None else lazy_cls
        self._lazy_args_ = args
        self._lazy_kwargs_ = kwargs
        # Lazy operates as partial by default, lazy_instance(*args,**kwargs) calls will start instance immediately
        self._lazy_default_keywords_ += ["__call__"] if lazy_partial else ""
        self._lazy_logger_.debug(
            f"{self._lazy_repr_()} Created    with {self._lazy_ref_cls_}"
        )
        self._lazy_logger_.debug(f"{self._lazy_repr_()}           {location_info()}")

    def _lazy_build_(self, *args, **kwargs):
        self._lazy_logger_.info(
            f"{self._lazy_repr_()} Starting   with args:{args}, kwargs:{kwargs}"
        )
        self._lazy_logger_.debug(f"{self._lazy_repr_()}           {location_info()}")
        self._lazy_ref_instance_ = self._lazy_builder_(*args, **kwargs)
        self._lazy_logger_.debug(f"{self._lazy_repr_()} Started   ")

    def _lazy_repr_(self):
        return f"<{self._lazy_cls_.__name__}({self._lazy_ref_cls_.__name__}) object at {hex(id(self))}>"

    def __repr__(self):
        if self._lazy_ref_instance_ is None:
            return self._lazy_repr_()
        return self._lazy_ref_instance_.__repr__()

    def __str__(self):
        if self._lazy_ref_instance_ is None:
            return self._lazy_repr_()
        return self._lazy_ref_instance_.__str__()

    def __format__(self, format_spec):
        if self._lazy_ref_instance_ is None:
            return self.__format__(format_spec)
        return self._lazy_ref_instance_.__format__(format_spec)

    def __call__(self, *args, **kwargs):
        if self._lazy_ref_instance_ is None:
            if self._lazy_args_ and args:
                raise SyntaxError(
                    f"Positional arguments already defined: {self._lazy_args_}"
                )
            use_args_ = args if args else self._lazy_args_
            use_kwargs_ = self._lazy_kwargs_.copy()
            use_kwargs_.update(kwargs)
            self._lazy_build_(*use_args_, **use_kwargs_)
            return self._lazy_instance_getter_

    @property
    def _lazy_instance_getter_(self) -> _T:
        if self._lazy_ref_instance_ is None:
            self._lazy_build_(*self._lazy_args_, **self._lazy_kwargs_)
        if self._lazy_ref_instance_ is None:
            raise NameError(f"Lazy builder {self._lazy_builder_} returned {None}")
        return self._lazy_ref_instance_

    def __getattribute__(self, item):
        logger.log(1, f"getattribute: {item}")
        if item.startswith("_lazy_") or item in self._lazy_reserved_keywords_:
            return super(Lazy, self).__getattribute__(item)
        if self._lazy_ref_instance_ is None and item in self._lazy_default_keywords_:
            return super(Lazy, self).__getattribute__(item)
        return self._lazy_instance_getter_.__getattribute__(item)

    def __setattr__(self, key: str, value):
        logger.log(2, f"setattr: {key}")
        if key.startswith("_lazy_") or key in self._lazy_reserved_keywords_:
            return super(Lazy, self).__setattr__(key, value)
        if self._lazy_ref_instance_ is None and key in self._lazy_default_keywords_:
            return super(Lazy, self).__setattr__(key, value)
        return self._lazy_instance_getter_.__setattr__(key, value)


for operator in supported_python_operators:
    setattr(Lazy, operator, proxy_operator(operator))


class LazyFactory(Generic[_T]):
    def __init__(
        self,
        lazy_cls: _T,
        *args,
        lazy_builder: Callable[[Any], _T] = None,
        lazy_logger: Logger = logger,
        **kwargs,
    ):
        self._lazy_ref_cls_ = lazy_cls
        self._lazy_builder_ = lazy_builder if lazy_builder is not None else lazy_cls
        self._lazy_args_ = args
        self._lazy_kwargs_ = kwargs
        self._lazy_logger_ = lazy_logger
        self._lazy_logger_.debug(f"{self} Created    with {self._lazy_ref_cls_}")
        self._lazy_logger_.debug(f"{self}           {location_info()}")

    def __call__(self, *args, **kwargs):
        if self._lazy_args_ and args:
            raise SyntaxError(
                f"Positional arguments already defined: {self._lazy_args_}"
            )
        use_args = args if args else self._lazy_args_
        use_kwargs = self._lazy_kwargs_.copy()
        use_kwargs |= kwargs
        use_kwargs |= {
            "lazy_cls": self._lazy_ref_cls_,
            "lazy_builder": self._lazy_builder_,
            "lazy_logger": self._lazy_logger_,
        }
        return Lazy(*use_args, **use_kwargs)


class LazyContext(Lazy, Generic[_T]):
    _lazy_reserved_keywords_ = ("__enter__", "__exit__")

    def _lazy_build_(self, *args, **kwargs) -> None:
        self._lazy_logger_.info(
            f"{self._lazy_repr_()} Starting   with args:{args}, kwargs:{kwargs}"
        )
        self._lazy_logger_.debug(f"          {location_info()}")
        self._lazy_ref_instance_ = self._lazy_builder_(*args, **kwargs).__enter__()  # type: ignore[operator]
        self._lazy_logger_.debug(f"{self._lazy_repr_()} Started   ")

    def __enter__(self) -> Union[Lazy, _T]:
        if self._lazy_ref_instance_ is None:
            return self
        return self._lazy_instance_getter_

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._lazy_ref_instance_ is not None:
            self._lazy_logger_.info(f"{self._lazy_repr_()} Exiting   ")
            self._lazy_logger_.debug(
                f"{self._lazy_repr_()}           {location_info()}"
            )
            return self._lazy_instance_getter_.__exit__(exc_type, exc_val, exc_tb)


def make_lazy(function: Union[Callable[[Any], _T], Any]):
    def wrapped(*args, **kwargs) -> Union[Lazy, _T]:
        use_kwargs = kwargs
        use_kwargs.update(
            dict(
                lazy_cls=function.__annotations__.get("return", MethodResponse),
                lazy_builder=function,
            )
        )
        return Lazy(
            *args,
            **kwargs,
        )

    return wrapped



import platform

from lazilypy.logs import logger


class MethodResponse:
    """
    Missing return type in wrapped function
    Using defaults.MethodResponse class type
    """

    pass


forced_operators = {
    str: {
        "__radd__": lambda self, *args, **kwargs: args[0] + self._lazy_instance_getter_
    }
}


def to_instance(item):
    if "lazilypy.lazy.Lazy" in str(type(item)):
        return item._lazy_instance_getter_
    return item


def proxy_operator(_lazy_operator_):
    def method(self, *args, **kwargs):
        logger.log(5, _lazy_operator_)
        logger.log(5, type(self))
        instance_args = (to_instance(arg) for arg in args)
        instance_kwargs = {key: to_instance(kwargs[key]) for key in kwargs}
        if (
            self._lazy_ref_cls_ in forced_operators
            and _lazy_operator_ in forced_operators[self._lazy_ref_cls_]
        ):
            return forced_operators[self._lazy_ref_cls_][_lazy_operator_](
                self, *instance_args, **instance_kwargs
            )
        return self._lazy_instance_getter_.__getattribute__(_lazy_operator_)(
            *instance_args, **instance_kwargs
        )

    return method


int_operators = [
    # Addition ( a + b )
    "__add__",
    "__radd__",
    # Division ( a / b )
    "__truediv__",
    "__rtruediv__",
    # Division ( a // b )
    "__floordiv__",
    "__rfloordiv__",
    # Bitwise And ( a & b )
    "__and___",
    "__rand__",
    # Bitwise Exclusive Or ( a ^ b )
    "__xor__",
    # "__rxor__", TODO: Not Supported
    # Bitwise Inversion ( ~ a )
    "__invert__",
    # Bitwise Or ( a | b )
    "__or__",
    "__ror__",
    # Exponentiation ( a ** b )
    "__pow__",
    "__rpow__",
    # Left Shift ( a << b )
    "__lshift__",
    "__rlshift__",
    # Modulo ( a % b )
    "__mod__",
    "__rmod__",
    # Multiplication ( a * b )
    "__mul__",
    "__rmul__",
    # Negation (Arithmetic) ( - a )
    "__neg__",
    # Right Shift ( a >> b )
    "__rshift__",
    "__rrshift__",
    # Subtraction ( a - b )
    "__sub__",
    "__rsub__",
    # Ordering ( a < b )
    "__lt__",
    # Ordering ( a <= b )
    "__le__",
    # Equality ( a == b )
    "__eq__",
    # Difference ( a != b )
    "__ne__",
    # Ordering ( a >= b )
    "__ge__",
    # Ordering ( a > b )
    "__gt__",
]

compatibility_builtins = [
    # math.ceil(a)
    "__ceil__",
    # math.floor(a)
    "__floor__",
]

numeric_builtins = [
    # abs(a_
    "__abs__",
    # int(a)
    "__int__",
    # float(a)
    "__float__",
    # bin(a)
    # "__bin__", TODO: Not Supported
]

supported_python_operators = (
    int_operators
    + numeric_builtins
    + (compatibility_builtins if platform.python_version() > "3.9.0" else [""])
    + [
        # Concatenation ( seq1 + seq2 )
        "__concat__",
        # Containment Test ( obj in seq )
        "__contains__",
        # Identity ( a is b )
        # Identity ( a is not b )
        # Both are implemented with memory references
        # Indexed Assignment ( obj[k] = v )
        "__setitem__",
        # Indexed Deletion ( del obj[k] )
        "__delitem__",
        # Indexing ( obj[k] )
        "__getitem__",
        # Matrix Multiplication ( a @ b )
        "__matmul__",
        # Negation (Logical) ( not a )
        "__not__",
        # Positive ( + a )
        "__pos__",
        # Slice Assignment ( seq[i:j] = values )
        "__setitem__",
        # Slice Deletion ( del seq[i:j] )
        "__delitem__",
        # Slicing ( seq[i:j] )
        "__getitem__",
        # String Formatting ( s % obj )
        "__mod__",
        # Truth Test ( obj )
        "__truth__",
    ]
)

old_python_operators = [
    # x and
    "__and__",
    # bool(
    "__bool__",
    "__cached__",
    "__concat__",
    # in x
    "__contains__",
    # delete x[y]
    "__delitem__",
    # x %
    "__divmod__",
    "__doc__",
    # x[y]
    "__getitem__",
    "__getnewargs__",
    # x >
    "__iadd__",
    "__iand__",
    "__iconcat__",
    "__ifloordiv__",
    "__ilshift__",
    "__imatmul__",
    "__imod__",
    "__imul__",
    "__index__",
    "__int__",
    "__inv__",
    "__invert__",
    "__ior__",
    "__ipow__",
    "__irshift__",
    "__isub__",
    "__itruediv__",
    "__ixor__",
    "__rdivmod__",  #
    "__sizeof__",  #
    "__trunc__",  #
]



import inspect
import logging


def get_logger(name="Lazy Logger"):
    logger = logging.Logger(name, level=30)
    __log_formatter = logging.Formatter(
        "%(levelname)-8s %(asctime)s: %(message)s", datefmt="%Y-%m-%d:%H:%M:%S"
    )
    __log_channel = logging.StreamHandler()
    __log_channel.setFormatter(__log_formatter)
    logger.addHandler(__log_channel)
    return logger


def location_info():
    frame = inspect.currentframe()
    while True:
        previous_frame = frame.f_back
        no_previous_frame = previous_frame is None
        no_globals = not hasattr(previous_frame, "f_globals")
        no_file_context = no_globals or "__file__" not in previous_frame.f_globals
        is_import_frame = (
            no_file_context or "importlib" in previous_frame.f_globals["__file__"]
        )
        if no_previous_frame or no_file_context or is_import_frame:
            break
        frame = frame.f_back

    return f"File \"{frame.f_globals['__file__']}\", line {frame.f_lineno}"


logger = get_logger()



from typing import TypeVar
from unittest.mock import MagicMock

from lazilypy import Lazy, logger

from lazilypy.defaults import supported_python_operators
from conftest import (
    get_lazy,
    lazy_created_and_not_started,
    lazy_builder_called_and_instance_started,
)


def test_addition_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int + 2 == 3
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 + lazy_int == 3
    lazy_builder_called_and_instance_started(lazy_int)


def test_concat_operator():
    lazy_one = get_lazy(str, "one")
    lazy_created_and_not_started(lazy_one)
    assert lazy_one + "two" == "onetwo"
    lazy_builder_called_and_instance_started(lazy_one)

    # TODO: Unsupported
    # lazy_two = get_lazy(str, "two")
    # print(type(lazy_two))
    # lazy_created_and_not_started(lazy_two)
    # assert "one" + lazy_two == "onetwo"
    # lazy_builder_called_and_instance_started(lazy_two)


def test_contains_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int in [2]) == (1 in [2])
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_list = get_lazy(list, [1, 2])
    lazy_created_and_not_started(lazy_list)
    print(lazy_list)
    assert (1 in lazy_list) == (1 in [1, 2])
    lazy_builder_called_and_instance_started(lazy_list)


def test_true_div_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int / 2 == 1 / 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 / lazy_int == 2 / 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_floor_div_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int // 2 == 1 // 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 // lazy_int == 2 // 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_bitwise_and_operator():
    # TODO: Unsupported
    # lazy_int: int | Lazy = get_lazy(int, 1)
    # lazy_created_and_not_started(lazy_int)
    # assert lazy_int & 2 == 1 & 2
    # lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 & lazy_int == 2 & 1
    assert 2 & lazy_int == 0
    lazy_builder_called_and_instance_started(lazy_int)


def test_bitwise_xor_operator():
    lazy_int: int | Lazy = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int ^ 2 == 1 ^ 2
    assert lazy_int ^ 2 == 3
    lazy_builder_called_and_instance_started(lazy_int)

    # TODO: Unsupported
    # lazy_int = get_lazy(int, 1)
    # lazy_created_and_not_started(lazy_int)
    # assert 2 ^ lazy_int == 2 ^ 1
    # assert 2 ^ lazy_int == 3
    # lazy_builder_called_and_instance_started(lazy_int)


def test_bitwise_inversion_operator():
    # TODO: Unsupported
    lazy_int: int | Lazy = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert ~lazy_int == ~1
    lazy_builder_called_and_instance_started(lazy_int)


def test_bitwise_or_operator():
    lazy_int: int | Lazy = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int | 2 == 1 | 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 | lazy_int == 2 | 1
    assert 2 | lazy_int == 3
    lazy_builder_called_and_instance_started(lazy_int)


def test_power_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int**2 == 1**2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2**lazy_int == 2**1
    lazy_builder_called_and_instance_started(lazy_int)


def test_floor_mod_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int % 2 == 1 % 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 % lazy_int == 2 % 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_multiplication_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int * 2 == 1 * 2
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 * lazy_int == 2 * 1
    lazy_builder_called_and_instance_started(lazy_int)


def test_greater_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int > 2) == (1 > 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 > lazy_int) == (2 > 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_lesser_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int < 2) == (1 < 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 < lazy_int) == (2 < 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_lesser_equal_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int <= 2) == (1 <= 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 <= lazy_int) == (2 <= 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_equal_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int == 2) == (1 == 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 == lazy_int) == (2 == 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_not_equal_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (lazy_int != 2) == (1 != 2)
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert (2 != lazy_int) == (2 != 1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_sub_operator():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert lazy_int - 2 == -1
    lazy_builder_called_and_instance_started(lazy_int)

    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert 2 - lazy_int == 1
    lazy_builder_called_and_instance_started(lazy_int)



import math

from conftest import (
    get_lazy,
    lazy_created_and_not_started,
    lazy_builder_called_and_instance_started,
)


def test_abs():
    lazy_int = get_lazy(int, -1)
    lazy_created_and_not_started(lazy_int)
    assert abs(lazy_int) == abs(-1)
    lazy_builder_called_and_instance_started(lazy_int)


def test_ceil():
    lazy_int = get_lazy(float, 1.3)
    lazy_created_and_not_started(lazy_int)
    assert math.ceil(lazy_int) == math.ceil(1.3)
    lazy_builder_called_and_instance_started(lazy_int)


def test_floor():
    lazy_int = get_lazy(float, 1.3)
    lazy_created_and_not_started(lazy_int)
    assert math.floor(lazy_int) == math.floor(1.3)
    lazy_builder_called_and_instance_started(lazy_int)


def test_int():
    lazy_int = get_lazy(float, 1.3)
    lazy_created_and_not_started(lazy_int)
    assert int(lazy_int) == int(1.3)
    lazy_builder_called_and_instance_started(lazy_int)


def test_float():
    lazy_int = get_lazy(int, 1)
    lazy_created_and_not_started(lazy_int)
    assert float(lazy_int) == float(1)
    lazy_builder_called_and_instance_started(lazy_int)
